/*
 * ConvRBM.cu
 *
 *  Created on: 27-Apr-2009
 *      Author: spoon
 */

#include <cutil_inline.h>
#include <time.h>
#include <matrix.h>
#include <nvmatrix.cuh>
#include <conv.cuh>
#include <conv3.cuh>
#include <conv_util.cuh>
#include <thread.h>
#include <queue.h>
#include <WorkRequest.h>
#include <WorkResult.h>

#include "ConvRBM.cuh"
#include "ConvRBM_kernel.cuh"

using namespace std;

ConvRBM::ConvRBM(Matrix* hHidVis, Matrix* hVisBiases, Matrix* hHidBiases,
                 Matrix* hHidVisInc, Matrix* hVisBiasInc, Matrix* hHidBiasInc,
                 Matrix* hHidMeans,
                 Matrix* hVisHidGlob, Matrix *hHidBiasesGlob,
                 Matrix* hVisHidIncGlob, Matrix* hHidBiasIncGlob,
                 Matrix* hHidMeansGlob,
                 Matrix* hNegDataPCD,
                 int fixedBorderSize, int localMaxSize, int zeroPadding,
                 float epsHidVis, float epsVisBias, float epsHidBias,
                 float epsHidVisGlob, float epsHidBiasGlob,
                 float mom, float weightCost,
                 float sparseCost, float sparseTarget,
                 float sparseCostGlob, float sparseTargetGlob,
                 int minibatchSize, int cd,
                 int deviceID) : Thread(false) {

    this->numVis                = hVisBiases->getNumCols();
    this->numHid                = hHidVis->getNumRows();
    this->filterSize            = int(sqrt(hHidVis->getNumCols() / 3));
    this->fixedBorderSize       = fixedBorderSize;
    this->imgSize               = int(sqrt(numVis / 3));
    this->paddedImgSize         = imgSize + 2 * zeroPadding;
    this->numConvOutputsX       = paddedImgSize - filterSize + 1;
    this->numConvOutputs        = numConvOutputsX * numConvOutputsX;
    this->imgCenterSize         = imgSize - 2 * fixedBorderSize;
    this->hidProbsPaddingSize   = filterSize - 1 - (fixedBorderSize + zeroPadding);
    this->paddedHidProbsSize    = numConvOutputsX + 2 * hidProbsPaddingSize;
    assert(hidProbsPaddingSize >= 0);

    this->numHidGlob            = hVisHidGlob->getNumCols();
    this->tiedBiases            = hHidBiases->getNumCols() == 1;

    this->localMaxSize          = localMaxSize;
    this->localMaxPixels        = localMaxSize * localMaxSize;
    this->zeroPadding           = zeroPadding;

    this->hHidVis               = hHidVis;
    this->hVisBiases            = hVisBiases;
    this->hHidBiases            = hHidBiases;
    this->hHidVisInc            = hHidVisInc;
    this->hVisBiasInc           = hVisBiasInc;
    this->hHidBiasInc           = hHidBiasInc;
    this->hVisHidGlob           = hVisHidGlob;
    this->hVisHidIncGlob        = hVisHidIncGlob;
    this->hHidBiasesGlob        = hHidBiasesGlob;
    this->hHidBiasIncGlob       = hHidBiasIncGlob;

    this->hHidMeans             = hHidMeans;
    this->hHidMeansGlob         = hHidMeansGlob;

    this->hNegDataPCD           = hNegDataPCD;

    this->epsHidVis             = epsHidVis;
    this->epsVisBias            = epsVisBias;
    this->epsHidBias            = tiedBiases ? epsHidBias / numConvOutputs : epsHidBias;
    this->epsHidVisGlob         = epsHidVisGlob;
    this->epsHidBiasGlob        = epsHidBiasGlob;
    this->mom                   = mom;
    this->weightCost            = weightCost;

    this->sparseCost            = sparseCost;
    this->sparseTarget          = sparseTarget;
    this->sparseCostGlob        = sparseCostGlob;
    this->sparseTargetGlob      = sparseTargetGlob;

    this->minibatchSize         = minibatchSize;
    this->cd                    = cd;

    this->requestQueue          = new Queue<WorkRequest*>();
    this->resultQueue           = new Queue<WorkResult*>();

    this->data                  = NULL;
    this->deviceID              = deviceID;
}

/*
 * There's really no need for a destructor since this object has the lifetime
 * of the entire application.
 */
ConvRBM::~ConvRBM() {

}

void ConvRBM::initCuda() {
    cudaSetDevice(deviceID < 0 ? cutGetMaxGflopsDeviceId() : deviceID);
    cublasInit();
    NVMatrix::initDeviceProps();
    NVMatrix::initRandom(time(0));
//    NVMatrix::initRandom(7);

    this->hidVis            = new NVMatrix(*hHidVis, true);
    this->visBiases         = new NVMatrix(*hVisBiases, true);
    this->hidBiases         = new NVMatrix(*hHidBiases, true);
    this->hidVisInc         = new NVMatrix(*hHidVisInc, true);
    this->visBiasInc        = new NVMatrix(*hVisBiasInc, true);
    this->hidBiasInc        = new NVMatrix(*hHidBiasInc, true);
    this->hidMeans          = new NVMatrix(*hHidMeans, true);

    this->visHidGlob        = new NVMatrix(*hVisHidGlob, true);
    this->visHidIncGlob     = new NVMatrix(*hVisHidIncGlob, true);
    this->hidBiasesGlob     = new NVMatrix(*hHidBiasesGlob, true);
    this->hidBiasIncGlob    = new NVMatrix(*hHidBiasIncGlob, true);
    this->hidMeansGlob      = new NVMatrix(*hHidMeansGlob, true);
    this->visHidGlob_T      = &visHidGlob->getTranspose();

    this->hidProbs          = new NVMatrix(false);
    this->hidProbsPadded    = new NVMatrix(false);
    this->posUnitProds      = new NVMatrix(false);
    this->negUnitProds      = new NVMatrix(false);
    this->unitProds_tmp     = new NVMatrix(false);
    this->posVisAct         = new NVMatrix(false);
    this->negVisAct         = new NVMatrix(false);
    this->posHidAct         = new NVMatrix(false);
    this->negHidAct         = new NVMatrix(false);
    this->hidProbs_tmp      = tiedBiases ? new NVMatrix() : NULL;
    this->hidBiases_tiled   = !tiedBiases ? new NVMatrix() : NULL;

    this->miniDataZeroPadded = zeroPadding > 0 ? new NVMatrix(false) : NULL;

    this->hidProbsGlob      = numHidGlob > 0 ? new NVMatrix() : NULL;
    this->posHidActGlob     = numHidGlob > 0 ? new NVMatrix() : NULL;
    this->negHidActGlob     = numHidGlob > 0 ? new NVMatrix() : NULL;
    this->negDataGlob       = numHidGlob > 0 ? new NVMatrix() : NULL;

    this->sparseGrads_tmp   = sparseCost > 0 ? new NVMatrix(false) : NULL;
    this->sparseBiasGrads   = sparseCost > 0 ? new NVMatrix(false) : NULL;
    this->sparseWeightGrads = sparseCost > 0 ? new NVMatrix(false) : NULL;

    this->sparseGradsGlob       = sparseCostGlob * numHidGlob > 0 ? new NVMatrix() : NULL;
    this->sparseWeightGradsGlob = sparseCostGlob * numHidGlob > 0 ? new NVMatrix() : NULL;
    this->sparseBiasGradsGlob   = sparseCostGlob * numHidGlob > 0 ? new NVMatrix() : NULL;

    this->negDataFull           = cd == 0 ? new NVMatrix(*hNegDataPCD, true) : new NVMatrix(false);
    this->negDataCenter         = fixedBorderSize > 0 ? new NVMatrix(false) : negDataFull;
    this->reconData             = cd == 0 ? new NVMatrix(*negDataFull, false) : negDataFull;

    this->localRegions_tmp      = localMaxSize > 0 ? new NVMatrix(false) : NULL;
    this->localMax_tmp          = localMaxSize > 0 ? new NVMatrix(false) : NULL;
    this->localSum_tmp          = localMaxSize > 0 ? new NVMatrix(false) : NULL;
    this->hidProbsSampled       = localMaxSize > 0 ? new NVMatrix(false) : hidProbs;
    this->sampleRandoms         = localMaxSize > 0 ? new NVMatrix(false) : NULL;

    this->data                  = new NVMatrix(); // The training data, initially null
}

void* ConvRBM::run() {
    /*
     * This gets executed in a new CPU thread so it's ok to initialize cuda stuff here.
     */
    initCuda();

    while (true) {
        WorkRequest* req = requestQueue->dequeue();

        if (req->getRequestType() == WorkRequest::TRAIN) {
            data->copyFromHost(*req->getHData(), true);
            train();
        } else if (req->getRequestType() == WorkRequest::TEST) {
            data->copyFromHost(*req->getHData(), true);
            test();
        } else if (req->getRequestType() == WorkRequest::SYNC) {
            syncWithHost();
        }

        delete req;
    }
    return NULL;
}

void ConvRBM::syncWithHost() {
    hidVis->copyToHost(*hHidVis);
    visBiases->copyToHost(*hVisBiases);
    hidBiases->copyToHost(*hHidBiases);
    hidVisInc->copyToHost(*hHidVisInc);
    visBiasInc->copyToHost(*hVisBiasInc);
    hidBiasInc->copyToHost(*hHidBiasInc);

    hidMeans->copyToHost(*hHidMeans);

    if(numHidGlob > 0) {
        visHidGlob->copyToHost(*hVisHidGlob);
        visHidIncGlob->copyToHost(*hVisHidIncGlob);
        hidBiasesGlob->copyToHost(*hHidBiasesGlob);
        hidBiasIncGlob->copyToHost(*hHidBiasIncGlob);
        hidMeansGlob->copyToHost(*hHidMeansGlob);
    }

    if (cd == 0) {
        negDataFull->copyToHost(*hNegDataPCD);
    }

    resultQueue->enqueue(new SyncDoneResult());
}

void ConvRBM::train() {
    int numMinibatches = DIVUP(data->getNumRows(), minibatchSize);
    double batchError = 0;
    for (int miniStart = 0; miniStart < data->getNumRows(); miniStart += minibatchSize) {
        int miniEnd = min(data->getNumRows(), miniStart + minibatchSize);
        NVMatrix* miniData = &data->sliceRows(miniStart, miniEnd); // view

        computeHidProbs(miniData);
        computeCDGrads(miniData, true);

        if(cd == 0) {
            computeHidProbs(negDataFull);
        }
        for (int c = 0; c < cd; c++) {
            computeNegData(miniData);
            computeHidProbs(negDataFull);
        }

        computeCDGrads(negDataFull, false);
        updateWeights();

        batchError += computeError(miniData); // overwrites reconData

        if (cd == 0) {
            computeNegData(miniData);
        }

        // All these matrices are views
        delete miniData;
    }
    cudaThreadSynchronize();
    resultQueue->enqueue(new BatchDoneResult(batchError / data->getNumElements()));
}

void ConvRBM::test() {
    int numMinibatches = DIVUP(data->getNumRows(), minibatchSize);
    double batchError = 0;
    for (int miniStart = 0; miniStart < data->getNumRows(); miniStart += minibatchSize) {
        int miniEnd = min(data->getNumRows(), miniStart + minibatchSize);
        NVMatrix* miniData = &data->sliceRows(miniStart, miniEnd); // view
        if (cd > 0) {
            computeHidProbs(miniData);

            for (int c = 0; c < cd; c++) {
                computeNegData(miniData);
                if (c < cd - 1)
                    computeHidProbs(negDataFull);
            }
        }
        batchError += computeError(miniData); // overwrites reconData

        // All these matrices are views
        delete miniData;
    }
    cudaThreadSynchronize();
    resultQueue->enqueue(new BatchDoneResult(batchError / data->getNumElements()));
}

inline void ConvRBM::computeHidProbs(NVMatrix* miniData) {
    int numCases = miniData->getNumRows();
    hidProbs->resize(numHid, numCases * numConvOutputs);
    if(zeroPadding == 0) {
        convolve(miniData, hidVis, hidProbs, 1, true);
    } else {
        miniDataZeroPadded->resize(numCases, paddedImgSize * paddedImgSize * 3);
        miniDataZeroPadded->apply(NVMatrix::ZERO);
        copyInto(miniData, miniDataZeroPadded, zeroPadding, true);
        convolve(miniDataZeroPadded, hidVis, hidProbs, 1, true);
    }

    if (!tiedBiases) {
        hidBiases->tile(1, numCases, *hidBiases_tiled);
        hidProbs->add(*hidBiases_tiled);
    } else {
        hidProbs->addVector(*hidBiases);
    }

    if (localMaxSize > 0) {
        /*
         * Compute hidden softmax activations.
         */
        localRegions_tmp->resize(hidProbs->getNumElements() / localMaxPixels, localMaxPixels);
        NVMatrix* hidProbsReshaped = &hidProbs->reshaped(numCases * numHid, numConvOutputs); // view
        gridToMatrix(hidProbsReshaped, localRegions_tmp, localMaxSize);

        localRegions_tmp->max(1, *localMax_tmp);
        localMax_tmp->scale(-1);
        localRegions_tmp->addVector(*localMax_tmp); // subtract max
        localRegions_tmp->apply(NVMatrix::EXP);
        localRegions_tmp->sum(1, *localSum_tmp);
        localMax_tmp->apply(NVMatrix::EXP); // e^-max
        localSum_tmp->add(*localMax_tmp); // add e^-max to sum to allow hidden units to be off
        localRegions_tmp->eltWiseDivideByVector(*localSum_tmp);  // divide by sum
        matrixToGrid(localRegions_tmp, hidProbsReshaped, localMaxSize);

        delete hidProbsReshaped;
    } else {
        hidProbs->apply(NVMatrix::LOGISTIC1);
//        dim3 blocks(DIVUP(hidProbs->getNumElements(), 256));
//        dim3 threads(256);
//        kZeroNegatives<<<blocks,threads>>>(hidProbs->getDevData(), hidProbs->getNumElements());
//        cutilCheckMsg("kZeroNegatives execution failed");
    }

    if (numHidGlob > 0) {
        miniData->rightMult(*visHidGlob, *hidProbsGlob);
        hidProbsGlob->addVector(*hidBiasesGlob);
        hidProbsGlob->apply(NVMatrix::LOGISTIC1);
    }
}

inline void ConvRBM::computeBiasGrads(NVMatrix* hidProbs, NVMatrix* target) {
    int numCases = hidProbs->getNumCols() / numConvOutputs;
    int threadsX = 16, threadsY = 16;
    int blocksX = DIVUP(numConvOutputs, threadsX);
    int blocksY = DIVUP(numHid, threadsY);
    int sumLength = numCases;
    target->resize(numHid, numConvOutputs);
    dim3 threads(threadsX, threadsY);
    dim3 blocks(blocksX, blocksY);
    kComputeBiasGrads<<<blocks, threads>>>(hidProbs->getDevData(), target->getDevData(), sumLength, numHid * numCases, numConvOutputs);
    cutilCheckMsg("kComputeBiasGrads execution failed");
}

inline void ConvRBM::computeCDGrads(NVMatrix* miniData, bool posPhase) {
    NVMatrix* unitProds = posPhase ? posUnitProds : negUnitProds;
    NVMatrix* visAct = posPhase ? posVisAct : negVisAct;
    NVMatrix* hidAct = posPhase ? posHidAct : negHidAct;
    int numCases = miniData->getNumRows();

    /*
     * Weight derivatives.
     */
    unitProds_tmp->resize(numCases, hidVis->getNumElements());
    convolve2(zeroPadding == 0 ? miniData : miniDataZeroPadded, hidProbs, unitProds_tmp, numConvOutputsX, 1, true);
    unitProds_tmp->sum(0, *unitProds);
    unitProds->reshape(numHid, filterSize * filterSize * 3);

    /*
     * Hidden bias derivatives.
     */
    if (!tiedBiases) {
        computeBiasGrads(hidProbs, hidAct);
        if (posPhase)
            hidMeans->add(*posHidAct, SPARSE_DAMPING, (1 - SPARSE_DAMPING) / numCases);
    } else {
        computeBiasGrads(hidProbs, hidProbs_tmp);
        if (posPhase)
            hidMeans->add(*hidProbs_tmp, SPARSE_DAMPING, (1 - SPARSE_DAMPING) / numCases);
//        hidProbs_tmp->reshape(numConvOutputs, numHid);
        hidProbs_tmp->sum(1, *hidAct);
//        hidAct->scale(1.0 / numConvOutputs);
    }

    if (posPhase) {
        if (sparseCost > 0) {
    //        posHidAct->reshape(1, posHidAct->getNumElements());

            dim3 blocks(DIVUP(hidProbs->getNumCols(), SPARSE_THREADS_X), DIVUP(hidProbs->getNumRows(), SPARSE_THREADS_Y));
            dim3 threads(SPARSE_THREADS_X, SPARSE_THREADS_Y);
            assert(blocks.x < NUM_BLOCKS_MAX && blocks.y < NUM_BLOCKS_MAX);
            sparseGrads_tmp->resize(*hidProbs);
            kComputeSparseGrads<<<blocks,threads>>>(hidProbs->getDevData(),hidMeans->getDevData(),
                                                    sparseGrads_tmp->getDevData(), sparseCost, sparseTarget,
                                                    hidProbs->getNumCols(), hidProbs->getNumRows(), numConvOutputs);
            cutilCheckMsg("kComputeSparseGrads execution failed");

            convolve2(zeroPadding == 0 ? miniData : miniDataZeroPadded, sparseGrads_tmp, unitProds_tmp, numConvOutputsX, 1, true);
            unitProds_tmp->sum(0, *sparseWeightGrads);
            sparseWeightGrads->reshape(numHid, filterSize * filterSize * 3);

            /*
             * Sparseness derivatives w.r.t. hidden biases.
             */
            if (!tiedBiases) {
                computeBiasGrads(sparseGrads_tmp, sparseBiasGrads);
            } else {
                computeBiasGrads(sparseGrads_tmp, hidProbs_tmp);
                hidProbs_tmp->sum(1, *sparseBiasGrads);
//                sparseBiasGrads->scale(1.0 / numConvOutputs);
            }
        }
    }

    if (numHidGlob > 0) {
        NVMatrix* miniData_T = &miniData->getTranspose();
        if (posPhase) {
            visHidIncGlob->addProduct(*miniData_T, *hidProbsGlob, mom, epsHidVisGlob / numCases);
        } else {
            visHidIncGlob->addProduct(*miniData_T, *hidProbsGlob, 1, -epsHidVisGlob / numCases);
        }

        hidProbsGlob->sum(0, posPhase ? *posHidActGlob : *negHidActGlob);

        /*
         * Sparseness derivatives w.r.t. global units.
         */
        if(posPhase) {
            hidMeansGlob->add(*posHidActGlob, SPARSE_DAMPING, (1 - SPARSE_DAMPING) / numCases);
            if (sparseCostGlob > 0) {

                assert(hidProbsGlob->isTrans()); // must be in col major order
                sparseGradsGlob->resize(*hidProbsGlob);
                kComputeSparseGradsGlob<<<SPARSE_BLOCKS_GLOB,SPARSE_THREADS_GLOB>>>(hidProbsGlob->getDevData(),hidMeansGlob->getDevData(),
                        sparseCostGlob, sparseTargetGlob, sparseGradsGlob->getDevData(), numCases, numHidGlob);
                cutilCheckMsg("kComputeSparseGradsGlob execution failed");
                sparseGradsGlob->sum(0, *sparseBiasGradsGlob);

                miniData_T->rightMult(*sparseGradsGlob, *sparseWeightGradsGlob);
            }
        }
        delete miniData_T;
    }

    /*
     * Visible bias derivatives.
     */
    miniData->sum(0, *visAct);
}

/*
 * Binarizes all hid probs. Needs posMiniData when fixedBorderSize > 0.
 */
inline void ConvRBM::computeNegData(NVMatrix* posMiniData) {
    int numCases = posMiniData->getNumRows();
    if (localMaxSize == 0) {
        hidProbsSampled->binarizeProbs(); // hidProbs == hidProbsSampled
    } else { // hidProbs != hidProbsSampled
        /*
         * Now sample from these softmaxes.
         */
        sampleRandoms->resize(localRegions_tmp->getNumRows(), 1);
        sampleRandoms->randomizeUniform();
        sampleMultinomial(localRegions_tmp, sampleRandoms, localRegions_tmp);
        hidProbsSampled->resize(*hidProbs);
        NVMatrix* hidProbsSampledReshaped = &hidProbsSampled->reshaped(numCases * numHid, numConvOutputs);
        matrixToGrid(localRegions_tmp, hidProbsSampledReshaped, localMaxSize);
        delete hidProbsSampledReshaped;
    }
    if (numHidGlob > 0) {
        hidProbsGlob->binarizeProbs();
    }

    negDataCenter->resize(numCases, imgCenterSize * imgCenterSize * 3);
    if (fixedBorderSize > 0) { // otherwise negDataCenter == negDataFull
        negDataFull->copyFromDevice(*posMiniData, true);
    }
    /*
     * Pad hidProbs with zeros.
     */
    hidProbsPadded->resize(numHid, numCases * paddedHidProbsSize * paddedHidProbsSize);
    hidProbsPadded->apply(NVMatrix::ZERO);

    NVMatrix* hidProbsReshaped = &hidProbsSampled->reshaped(numCases * numHid, numConvOutputs); // view
    copyInto(hidProbsReshaped, hidProbsPadded, hidProbsPaddingSize, false);
    delete hidProbsReshaped;

    /*
     * Negative data.
     */
    convolve3(hidProbsPadded, hidVis, negDataCenter, 1, true);
    if (numHidGlob > 0) {
        hidProbsGlob->rightMult(*visHidGlob_T, *negDataGlob);
        negDataCenter->add(*negDataGlob);
    }
    if (fixedBorderSize > 0) {
        copyInto(negDataCenter, negDataFull, fixedBorderSize, true);
    }
    negDataFull->addVector(*visBiases);
}

inline double ConvRBM::computeError(NVMatrix* pos) {
    negDataFull->squaredDiff(*pos, *reconData);
    return reconData->sum();
}

inline void ConvRBM::updateWeights() {
    int numCases = negDataFull->getNumRows();

    /*
     * Seriously no need to do momentum on these wight updates which are already
     * an average over thousands of locations.
     */
    hidVisInc->addSum(*posUnitProds, *negUnitProds, 0, epsHidVis / numCases , -epsHidVis / numCases );
    if (!WEIGHT_NORMALIZATION) {
        hidVisInc->add(*hidVis, -epsHidVis * weightCost);
    }
    hidBiasInc->addSum(*posHidAct, *negHidAct, mom, epsHidBias / numCases, -epsHidBias / numCases);
    visBiasInc->addSum(*posVisAct, *negVisAct, mom, epsVisBias / numCases, -epsVisBias / numCases);

    hidVis->add(*hidVisInc);
    hidBiases->add(*hidBiasInc);
    visBiases->add(*visBiasInc);

    if (sparseCost > 0) { // no momentum on these updates
        hidBiases->add(*sparseBiasGrads, -epsHidBias / numCases);
        hidVis->add(*sparseWeightGrads, -epsHidVis / numCases);
    }

    if (numHidGlob > 0) {
        hidBiasIncGlob->addSum(*posHidActGlob, *negHidActGlob, mom, epsHidBiasGlob / numCases, -epsHidBiasGlob / numCases);
        visHidIncGlob->add(*visHidGlob, -weightCost * epsHidVisGlob / numCases);

        visHidGlob->add(*visHidIncGlob);
        hidBiasesGlob->add(*hidBiasIncGlob);

        if (sparseCostGlob > 0) {
            hidBiasesGlob->add(*sparseBiasGradsGlob, -epsHidBiasGlob / numCases);
            visHidGlob->add(*sparseWeightGradsGlob, -epsHidVisGlob / numCases);
        }
    }

    if (WEIGHT_NORMALIZATION) {
        hidVis->apply(NVMatrix::SQUARE, *negUnitProds);
        NVMatrix* sum = &negUnitProds->sum(1);
        sum->apply(NVMatrix::SQRT);
        dim3 blocks(DIVUP(hidVis->getNumCols(), WEIGHT_NORM_THREADS_X), numHid);
        dim3 threads(WEIGHT_NORM_THREADS_X, 1);
        kWeightNorm<<<blocks, threads>>>(hidVis->getDevData(), sum->getDevData(), hidVis->getNumRows(), hidVis->getNumCols(), weightCost);
        cutilCheckMsg("kWeightNorm execution failed");
        cudaThreadSynchronize();
        delete sum;
    }
}
