/*
 * testnvmatrix.cu
 *
 *  Created on: Dec 13, 2009
 *      Author: Alex Krizhevsky (akrizhevsky@gmail.com)
 */

#include <cutil_inline.h>
#include <assert.h>
#include <matrix.h>
#include "nvmatrix.cuh"
#include "gpu_locking.h"

static unsigned int timer;

void init_tests(int boardNum) {
    cudaSetDevice(boardNum > -1 ? boardNum : cutGetMaxGflopsDeviceId());
    cublasInit();
    NVMatrix::initDeviceProps();
    NVMatrix::initRandom(7);
    cutilCheckError( cutCreateTimer( &timer));
}

void test_hardTranspose(int height, int width) {
    printf("===============================\n");
    printf("test_hardTranspose\n");
    printf("===============================\n");

    Matrix mat(height, width);
    mat.randomizeUniform();

    printf("Height: %d\n", height);
    printf("Width: %d\n", width);

    cutilCheckError( cutResetTimer( timer));
    cutilCheckError( cutStartTimer( timer));


    cutilCheckError( cutStopTimer( timer));
    printf("CPU (partial) result:\n");
    mat.print(0, 3, 0, 8);
    printf("CPU time: %.6f msec\n", cutGetTimerValue(timer));

    NVMatrix nvMat(mat, true);

    cutilCheckError( cutResetTimer( timer));
    cutilCheckError( cutStartTimer( timer));
    nvMat.flipTrans();

    cudaThreadSynchronize();
    cutilCheckError( cutStopTimer( timer));
    printf("GPU (partial) result:\n");
    nvMat.print(0, 3, 0, 8);
    printf("GPU time: %.6f msec\n", cutGetTimerValue(timer));

    // Compare results
    Matrix cpuNVTargets(mat);
    nvMat.copyToHost(cpuNVTargets);
    cpuNVTargets.subtract(mat);
    cpuNVTargets.apply(Matrix::ABS);
    printf("Max diff between CPU/GPU: %.6f\n", cpuNVTargets.max());
}

void test_rowSum(int height, int width) {
    printf("===============================\n");
    printf("test_rowSum\n");
    printf("===============================\n");
//    int width = 128*64;
//    int height = 1024;
    Matrix multi(height, width);
    Matrix targets(height, 1);

    multi.randomizeUniform();
    targets.apply(Matrix::ZERO);

    NVMatrix nvMulti(multi, true);
    NVMatrix nvTargets(targets, true);

    printf("Height: %d\n", height);
    printf("Width: %d\n", width);

    cutilCheckError( cutResetTimer( timer));
    cutilCheckError( cutStartTimer( timer));

    multi.sum(1, targets);

    cutilCheckError( cutStopTimer( timer));
    printf("CPU (partial) result:\n");
    targets.print(0, 3, 0, 8);
    printf("CPU time: %.6f msec\n", cutGetTimerValue(timer));

    cutilCheckError( cutResetTimer( timer));
    cutilCheckError( cutStartTimer( timer));
    nvMulti.sum(1, nvTargets);

    cudaThreadSynchronize();
    cutilCheckError( cutStopTimer( timer));
    printf("GPU (partial) result:\n");
    nvTargets.print(0, 3, 0, 8);
    printf("GPU time: %.6f msec\n", cutGetTimerValue(timer));

    // Compare results
    Matrix cpuNVTargets(targets);
    nvTargets.copyToHost(cpuNVTargets);
    cpuNVTargets.subtract(targets);
    cpuNVTargets.apply(Matrix::ABS);
    printf("Max diff between CPU/GPU: %.6f\n", cpuNVTargets.max());
}

void test_rowSumTrans(int height, int width) {
    printf("===============================\n");
    printf("test_rowSumTrans\n");
    printf("===============================\n");
//    int width = 128*64;
//    int height = 1024;
    Matrix multi(height, width);
    Matrix &multiT = multi.transpose();
    Matrix targets(1, height);

    multi.randomizeUniform();
    targets.apply(Matrix::ZERO);

    NVMatrix nvMultiT(multiT, true);
    NVMatrix nvTargets(targets, true);

    printf("Height: %d\n", height);
    printf("Width: %d\n", width);

    cutilCheckError( cutResetTimer( timer));
    cutilCheckError( cutStartTimer( timer));

    multiT.sum(0, targets);

    cutilCheckError( cutStopTimer( timer));
    printf("CPU (partial) result:\n");
    targets.print(0, 3, 0, 8);
    printf("CPU time: %.6f msec\n", cutGetTimerValue(timer));

    cutilCheckError( cutResetTimer( timer));
    cutilCheckError( cutStartTimer( timer));
    nvMultiT.sum(0, nvTargets);

    cudaThreadSynchronize();
    cutilCheckError( cutStopTimer( timer));
    printf("GPU (partial) result:\n");
    nvTargets.print(0, 3, 0, 8);
    printf("GPU time: %.6f msec\n", cutGetTimerValue(timer));

    // Compare results
    Matrix cpuNVTargets(targets);
    nvTargets.copyToHost(cpuNVTargets);
    cpuNVTargets.subtract(targets);
    cpuNVTargets.apply(Matrix::ABS);
    printf("Max diff between CPU/GPU: %.6f\n", cpuNVTargets.max());
}

void test_colSum(int height, int width) {
    printf("===============================\n");
    printf("test_colSum\n");
    printf("===============================\n");
//    int width = 128*64;
//    int height = 1024;
    Matrix multi(height, width);
    Matrix targets(1, width);

    multi.randomizeUniform();
    targets.apply(Matrix::ZERO);

    NVMatrix nvMulti(multi, true);
    NVMatrix nvTargets(targets, true);

    printf("Height: %d\n", height);
    printf("Width: %d\n", width);

    cutilCheckError( cutResetTimer( timer));
    cutilCheckError( cutStartTimer( timer));

    multi.sum(0, targets);

    cutilCheckError( cutStopTimer( timer));
    printf("CPU (partial) result:\n");
    targets.print(0, 3, 0, 8);
    printf("CPU time: %.6f msec\n", cutGetTimerValue(timer));

    cutilCheckError( cutResetTimer( timer));
    cutilCheckError( cutStartTimer( timer));
    nvMulti.sum(0, nvTargets);

    cudaThreadSynchronize();
    cutilCheckError( cutStopTimer( timer));
    printf("GPU (partial) result:\n");
    nvTargets.print(0, 3, 0, 8);
    printf("GPU time: %.6f msec\n", cutGetTimerValue(timer));

    // Compare results
    Matrix cpuNVTargets(targets);
    nvTargets.copyToHost(cpuNVTargets);
    cpuNVTargets.subtract(targets);
    cpuNVTargets.apply(Matrix::ABS);
    printf("Max diff between CPU/GPU: %.6f\n", cpuNVTargets.max());
}


void test_addTrans3(int height, int width) {
    printf("===============================\n");
    printf("test_addTrans3\n");
    printf("===============================\n");

    printf("Height: %d\n", height);
    printf("Width: %d\n", width);

    Matrix mat(height, width);
    Matrix mat2(height, width);
    Matrix mat3(height, width);

    NVMatrix nvMat(height, width, true);
    NVMatrix nvMat2(height, width, false);
    NVMatrix nvMat3(height, width, false);

    nvMat.randomizeUniform();
    nvMat2.randomizeUniform();
    nvMat3.randomizeUniform();

    nvMat.copyToHost(mat);
    nvMat2.copyToHost(mat2);
    nvMat3.copyToHost(mat3);

    cutilCheckError( cutResetTimer( timer));
    cutilCheckError( cutStartTimer( timer));
    mat.add(mat2);
    mat.add(mat3);
    cutilCheckError( cutStopTimer( timer));
    printf("CPU (partial) result:\n");
    mat.print(0, 3, 0, 8);
    printf("CPU time: %.6f msec\n", cutGetTimerValue(timer));

    cutilCheckError( cutResetTimer( timer));
    cutilCheckError( cutStartTimer( timer));
    nvMat.addSum(nvMat2, nvMat3, 1, 1, 1);

    cudaThreadSynchronize();
    cutilCheckError( cutStopTimer( timer));
    printf("GPU (partial) result:\n");
    nvMat.print(0, 3, 0, 8);
    printf("GPU time: %.6f msec\n", cutGetTimerValue(timer));

    // Compare results
    Matrix cpuNVTargets(mat);
    nvMat.copyToHost(cpuNVTargets);
    cpuNVTargets.subtract(mat);
    cpuNVTargets.apply(Matrix::ABS);
    printf("Max diff between CPU/GPU: %.6f\n", cpuNVTargets.max());
}

void test_addTrans(int height, int width) {
    printf("===============================\n");
    printf("test_addTrans\n");
    printf("===============================\n");

    printf("Height: %d\n", height);
    printf("Width: %d\n", width);

    Matrix mat(height, width);
    Matrix mat2(height, width);

    NVMatrix nvMat(height, width, true);
    NVMatrix nvMat2(height, width, false);

    nvMat.randomizeUniform();
    nvMat2.randomizeUniform();

    nvMat.copyToHost(mat);
    nvMat2.copyToHost(mat2);

    cutilCheckError( cutResetTimer( timer));
    cutilCheckError( cutStartTimer( timer));
    mat.add(mat2);
    cutilCheckError( cutStopTimer( timer));
    printf("CPU (partial) result:\n");
    mat.print(0, 3, 0, 8);
    printf("CPU time: %.6f msec\n", cutGetTimerValue(timer));

    cutilCheckError( cutResetTimer( timer));
    cutilCheckError( cutStartTimer( timer));
    nvMat.add(nvMat2);

    cudaThreadSynchronize();
    cutilCheckError( cutStopTimer( timer));
    printf("GPU (partial) result:\n");
    nvMat.print(0, 3, 0, 8);
    printf("GPU time: %.6f msec\n", cutGetTimerValue(timer));

    // Compare results
    Matrix cpuNVTargets(mat);
    nvMat.copyToHost(cpuNVTargets);
    cpuNVTargets.subtract(mat);
    cpuNVTargets.apply(Matrix::ABS);
    printf("Max diff between CPU/GPU: %.6f\n", cpuNVTargets.max());
}

void test_squaredDiffTrans(int height, int width) {
    printf("===============================\n");
    printf("test_squaredDiffTrans\n");
    printf("===============================\n");

    printf("Height: %d\n", height);
    printf("Width: %d\n", width);

    Matrix mat(height, width);
    Matrix mat2(height, width);

    NVMatrix nvMat(height, width, true);
    NVMatrix nvMat2(height, width, false);

    nvMat.randomizeUniform();
    nvMat2.randomizeUniform();

    nvMat.copyToHost(mat);
    nvMat2.copyToHost(mat2);

    cutilCheckError( cutResetTimer( timer));
    cutilCheckError( cutStartTimer( timer));
    mat.subtract(mat2);
    mat.apply(Matrix::SQUARE);
    cutilCheckError( cutStopTimer( timer));
    printf("CPU (partial) result:\n");
    mat.print(0, 3, 0, 8);
    printf("CPU time: %.6f msec\n", cutGetTimerValue(timer));

    cutilCheckError( cutResetTimer( timer));
    cutilCheckError( cutStartTimer( timer));
    nvMat.squaredDiff(nvMat2);

    cudaThreadSynchronize();
    cutilCheckError( cutStopTimer( timer));
    printf("GPU (partial) result:\n");
    nvMat.print(0, 3, 0, 8);
    printf("GPU time: %.6f msec\n", cutGetTimerValue(timer));

    // Compare results
    Matrix cpuNVTargets(mat);
    nvMat.copyToHost(cpuNVTargets);
    cpuNVTargets.subtract(mat);
    cpuNVTargets.apply(Matrix::ABS);
    printf("Max diff between CPU/GPU: %.6f\n", cpuNVTargets.max());
}

void test_eltWiseMultTrans(int height, int width) {
    printf("===============================\n");
    printf("test_eltWiseMultTrans\n");
    printf("===============================\n");

    printf("Height: %d\n", height);
    printf("Width: %d\n", width);

    Matrix mat(height, width);
    Matrix mat2(height, width);

    NVMatrix nvMat(height, width, true);
    NVMatrix nvMat2(height, width, false);

    nvMat.randomizeUniform();
    nvMat2.randomizeUniform();

    nvMat.copyToHost(mat);
    nvMat2.copyToHost(mat2);

    cutilCheckError( cutResetTimer( timer));
    cutilCheckError( cutStartTimer( timer));
    mat.eltWiseMult(mat2);
    cutilCheckError( cutStopTimer( timer));
    printf("CPU (partial) result:\n");
    mat.print(0, 3, 0, 8);
    printf("CPU time: %.6f msec\n", cutGetTimerValue(timer));

    cutilCheckError( cutResetTimer( timer));
    cutilCheckError( cutStartTimer( timer));
    nvMat.eltWiseMult(nvMat2);

    cudaThreadSynchronize();
    cutilCheckError( cutStopTimer( timer));
    printf("GPU (partial) result:\n");
    nvMat.print(0, 3, 0, 8);
    printf("GPU time: %.6f msec\n", cutGetTimerValue(timer));

    // Compare results
    Matrix cpuNVTargets(mat);
    nvMat.copyToHost(cpuNVTargets);
    cpuNVTargets.subtract(mat);
    cpuNVTargets.apply(Matrix::ABS);
    printf("Max diff between CPU/GPU: %.6f\n", cpuNVTargets.max());
}


void test_eltWiseDivideTrans(int height, int width) {
    printf("===============================\n");
    printf("test_eltWiseDivideTrans\n");
    printf("===============================\n");

    printf("Height: %d\n", height);
    printf("Width: %d\n", width);

    Matrix mat(height, width);
    Matrix mat2(height, width);

    NVMatrix nvMat(height, width, true);
    NVMatrix nvMat2(height, width, false);

    nvMat.randomizeUniform();
    nvMat2.randomizeUniform();

    nvMat.copyToHost(mat);
    nvMat2.copyToHost(mat2);

    cutilCheckError( cutResetTimer( timer));
    cutilCheckError( cutStartTimer( timer));
    mat.eltWiseDivide(mat2);
    cutilCheckError( cutStopTimer( timer));
    printf("CPU (partial) result:\n");
    mat.print(0, 3, 0, 8);
    printf("Max divide result: %.6f\n", std::max(-mat.min(), mat.max()));
    printf("CPU time: %.6f msec\n", cutGetTimerValue(timer));

    cutilCheckError( cutResetTimer( timer));
    cutilCheckError( cutStartTimer( timer));
    nvMat.eltWiseDivide(nvMat2);

    cudaThreadSynchronize();
    cutilCheckError( cutStopTimer( timer));
    printf("GPU (partial) result:\n");
    nvMat.print(0, 3, 0, 8);
    printf("GPU time: %.6f msec\n", cutGetTimerValue(timer));

    // Compare results
    Matrix cpuNVTargets(mat);
    nvMat.copyToHost(cpuNVTargets);
    cpuNVTargets.subtract(mat);
    cpuNVTargets.apply(Matrix::ABS);
    printf("Max diff between CPU/GPU: %.6f\n", cpuNVTargets.max());
}

void test_equalsVector(int height, int width, bool colVec) {
    printf("===============================\n");
    printf("test_equalsVector\n");
    printf("===============================\n");

    printf("Height: %d\n", height);
    printf("Width: %d\n", width);

    Matrix mat(height, width);
    Matrix vec(colVec ? height : 1, colVec ? 1 : width);

    NVMatrix nvMat(height, width);
    NVMatrix nvVec(colVec ? height : 1, colVec ? 1 : width);

    nvMat.randomizeUniform();
    nvMat.max(colVec ? 1 : 0, nvVec);

    nvMat.copyToHost(mat);
    nvVec.copyToHost(vec);
    Matrix& vecTiled = vec.tile(colVec ? 1 : height, colVec ? width : 1);

    cutilCheckError( cutResetTimer( timer));
    cutilCheckError( cutStartTimer( timer));
    mat.equals(vecTiled);
    cutilCheckError( cutStopTimer( timer));
    printf("CPU (partial) result:\n");
    mat.print(0, 3, 0, 8);
    printf("CPU time: %.6f msec\n", cutGetTimerValue(timer));

    cutilCheckError( cutResetTimer( timer));
    cutilCheckError( cutStartTimer( timer));
    nvMat.equalsVector(nvVec);

    cudaThreadSynchronize();
    cutilCheckError( cutStopTimer( timer));
    printf("GPU (partial) result:\n");
    nvMat.print(0, 3, 0, 8);
    printf("GPU time: %.6f msec\n", cutGetTimerValue(timer));

    // Compare results
    Matrix cpuNVTargets(mat);
    nvMat.copyToHost(cpuNVTargets);
    cpuNVTargets.subtract(mat);
    cpuNVTargets.apply(Matrix::ABS);
    printf("Max diff between CPU/GPU: %.6f\n", cpuNVTargets.max());
}

void test_sliceRows(int height, int width) {
    printf("===============================\n");
    printf("test_sliceRows\n");
    printf("===============================\n");

    printf("Height: %d\n", height);
    printf("Width: %d\n", width);

    Matrix mat(height, width);
    Matrix slice(2, width);

    NVMatrix nvMat(height, width);
    NVMatrix nvSlice(2, width);

    nvMat.randomizeUniform();
    nvMat.copyToHost(mat);

    cutilCheckError( cutResetTimer( timer));
    cutilCheckError( cutStartTimer( timer));
    mat.sliceRows(0, 2, slice);
    cutilCheckError( cutStopTimer( timer));
    printf("CPU (partial) result:\n");
    slice.print(0, 2, 0, 8);
    printf("CPU time: %.6f msec\n", cutGetTimerValue(timer));

    cutilCheckError( cutResetTimer( timer));
    cutilCheckError( cutStartTimer( timer));
    nvMat.sliceRows(0, 2, nvSlice);

    cudaThreadSynchronize();
    cutilCheckError( cutStopTimer( timer));
    printf("GPU (partial) result:\n");
    nvSlice.print(0, 2, 0, 8);
    printf("GPU time: %.6f msec\n", cutGetTimerValue(timer));

    // Compare results
    Matrix cpuNVSlice(slice);
    nvSlice.copyToHost(cpuNVSlice);
    cpuNVSlice.subtract(slice);
    cpuNVSlice.apply(Matrix::ABS);
    printf("Max diff between CPU/GPU: %.6f\n", cpuNVSlice.max());
}

void test_copy(int height, int width, int srcStartRow, int srcStartCol, int destStartRow,
                    int destStartCol, int numRows, int numCols, bool transSrc, bool transDest) {
    printf("===============================\n");
    printf("test_copy\n");
    printf("===============================\n");

    printf("Height: %d\n", height);
    printf("Width: %d\n", width);
    printf("srcStartRow: %d\n", srcStartRow);
    printf("srcStartCol: %d\n", srcStartCol);
    printf("destStartRow: %d\n", destStartRow);
    printf("destStartCol: %d\n", destStartCol);
    printf("numRows: %d\n", numRows);
    printf("numCols: %d\n", numCols);
    printf("transSrc: %d\n", transSrc);
    printf("transDest: %d\n", transDest);

    Matrix mat(height, width);
    Matrix mat2(height+77, width+55);

    NVMatrix nvMat(height, width, transSrc);
    NVMatrix nvMat2(height+77, width+55, transDest);

    nvMat.randomizeUniform();
    nvMat2.randomizeUniform();
    nvMat.copyToHost(mat);
    nvMat2.copyToHost(mat2);

    cutilCheckError( cutResetTimer( timer));
    cutilCheckError( cutStartTimer( timer));
    mat.copy(mat2, srcStartRow, srcStartRow + numRows, srcStartCol, srcStartCol + numCols, destStartRow, destStartCol);
    cutilCheckError( cutStopTimer( timer));
    printf("CPU (partial) result:\n");
    mat2.print(0, 4, 0, 8);
    printf("CPU time: %.6f msec\n", cutGetTimerValue(timer));

    cutilCheckError( cutResetTimer( timer));
    cutilCheckError( cutStartTimer( timer));
    nvMat.copy(nvMat2, srcStartRow, srcStartRow + numRows, srcStartCol, srcStartCol + numCols, destStartRow, destStartCol);

    cudaThreadSynchronize();
    cutilCheckError( cutStopTimer( timer));
    printf("GPU (partial) result:\n");
    nvMat2.print(0, 4, 0, 8);
    printf("GPU time: %.6f msec\n", cutGetTimerValue(timer));

    // Compare results
    Matrix cpuNVMat2(mat2);
    nvMat2.copyToHost(cpuNVMat2);
    cpuNVMat2.subtract(mat2);
    cpuNVMat2.apply(Matrix::ABS);
    printf("Max diff between CPU/GPU: %.6f\n", cpuNVMat2.max());
}

int main(int argc, char** argv) {
    // This line just for compiling and examining profiler output.
//    exit(0); conv3_bw_nofit_16x16<true,true,1><<<1,1>>>(NULL, NULL, NULL, 0,0, 0);
    int boardNum = get_board_lock();
    if (boardNum == GPU_LOCK_NO_BOARD) {
        printf("No free GPU boards!\n");
        exit(EXIT_FAILURE);
    } else if(boardNum == GPU_LOCK_NO_SCRIPT) {
        printf("Running on default board.\n");
    } else {
        printf("Running on board %d\n", boardNum);
    }

    init_tests(boardNum);

//    test_hardTranspose(4096+5, 1024+9);
//    test_addTrans3(1300000, 2);
//    test_addTrans(128, 3072);
//    test_equalsVector(3071,1000, false);
//    test_colSum(128, 48*24*24);
//    test_sliceRows(8, 8);
    test_copy(500, 400, 111, 20, 155, 233, 29, 121, true, true);
//    test_copy(500, 400, 0, 0, 0, 0, 2, 8, true, true);
}
